﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Diagnostics;
using System.Collections.Concurrent;

using GrapeCity.Documents.Pdf.Spec;
using GrapeCity.Documents.Pdf.ViewerSupportApi.Models;
using GrapeCity.Documents.Pdf.Annotations;
using GrapeCity.Documents.Pdf.AcroForms;
using GrapeCity.Documents.Pdf.Layers;
using GrapeCity.Documents.Pdf.ViewerSupportApi.Collaboration;
using GrapeCity.Documents.Pdf.ViewerSupportApi.Controllers;
using GrapeCity.Documents.Pdf.ViewerSupportApi.Connection;
using GrapeCity.Documents.Pdf.Security;

using _Float = System.Single;
using _Size = System.Drawing.SizeF;
using GrapeCity.Documents.Imaging;
using GrapeCity.Documents.Common;
using System.Text.RegularExpressions;
using GrapeCity.Documents.Pdf.Actions;

namespace GrapeCity.Documents.Pdf.ViewerSupportApi.Utils
{
    /// <summary>
    /// PDF document loader.
    /// </summary>
    public class GcPdfDocumentLoader : IDisposable
    {
        #region ** fields

        private static object lockDocumentLoad = new object();
        private _Float _resolution;

        private GcPdfDocument _doc;
        private SupportApiSettings _settings;
        private DocumentOptions _docOptions;
        private readonly SharedAccessMode _accessMode;
        private string _documentId;

        #endregion

        #region ** constructor

        /// <summary>
        /// Constructor.
        /// </summary>
        public GcPdfDocumentLoader(SupportApiSettings settings, DocumentOptions docOptions, SharedAccessMode accessMode, _Float resolution = 96)
        {
            _settings = settings;
            _docOptions = docOptions;
            _resolution = resolution;
            Url = "";
            ClientId = docOptions.clientId;
            _accessMode = accessMode;
        }

        #endregion

        #region ** properties

        /// <summary>
        /// Gets currently opened document.
        /// </summary>
        public GcPdfDocument Document
        {
            get
            {
                return _doc;
            }
        }

        public Stream Stream { get; private set; }

        public string FileName
        {
            get
            {
                string fileName = "";
                if (Info != null)
                    fileName = Info.fileName;
                return fileName;
            }
        }


        public string ClientId { get; private set; }

        public OpenDocumentInfo Info { get; private set; }

        public string Url { get; }

        public int Rotate { get; set; } = 0;

        public _Float Scale { get; set; } = 1.0f;

        public bool RenderInteractiveForms { get; internal set; }

        /// <summary>
        /// Document modifications.
        /// </summary>
        public DocumentModifications DocumentModifications { get; private set; }

        public string DocumentId
        {
            get
            {
                return _documentId;
            }
            set
            {
#if DEBUG
                Debug.Assert(string.IsNullOrEmpty(value) || string.IsNullOrEmpty(_documentId) || _documentId.Equals(value), "Error. DocumentId cannot be changed.");
#endif
                if (!string.IsNullOrEmpty(value))
                {
                    var clientConnection = ClientConnection.GetByClientId(value);
                    if (clientConnection != null)
                    {
                        clientConnection.DocumentId = value;
                    }
                }
                _documentId = value;
            }
        }

        #endregion

        #region ** methods

        public virtual void ApplyDocumentModifications(DocumentModifications modifications, string correlationId)
        {
            DocumentModifications = modifications;
            _applyModifications(correlationId);
        }

        public virtual OpenDocumentInfo Open(Stream data, string knownDocumentId = null, string knownFileName = null)
        {
            lock (lockDocumentLoad)
            {
                if (Stream != null)
                {
                    Stream.Dispose();
                    Stream = null;
                    _doc = null;
                }
                Stream = data;
                return _OpenInitialDocState(knownDocumentId, knownFileName);
            }
        }

        public void Reset()
        {
            DocumentModifications = null;
        }


        /// <summary>
        /// Returns value indicating whether the document is signed.
        /// </summary>
        public virtual bool IsSigned
        {
            get
            {
                var doc = _doc;
                if (doc == null)
                    return false;
                return SignatureFields.Where(f => f.Value != null).Count() > 0;
            }
        }

        /// <summary>
        /// Returns all available signature fields in document.
        /// </summary>
        public virtual List<SignatureField> SignatureFields
        {
            get
            {
                List<SignatureField> result = new List<SignatureField>();
                var doc = _doc;
                if (doc == null)
                    return result;
                foreach (var field in doc.AcroForm.Fields)
                {
                    if (field is SignatureField)
                        result.Add((SignatureField)field);
                }
                return result;
            }
        }

        /// <summary>
        /// Defines the modes of saving a PDF.
        /// </summary>
        public virtual SaveMode GetSaveMode(string correlationId)
        {

            Nullable<SaveMode> saveMode = null;
            var _docOptions = GcPdfViewerController.FindDocumentOptions(ClientId, correlationId);
            var saveSettings = _docOptions != null ? _docOptions.saveSettings : null;
            if (saveSettings != null && !string.IsNullOrEmpty(saveSettings.saveMode))
            {
                switch (saveSettings.saveMode)
                {
                    case "Default":
                        saveMode = SaveMode.Default;
                        break;
                    case "Linearized":
                        saveMode = SaveMode.Linearized;
                        break;
                    case "IncrementalUpdate":
                        saveMode = SaveMode.IncrementalUpdate;
                        break;
                    case "Auto":
                        break;
                    default:
                        saveMode = GcPdfViewerController.Settings.SaveMode;
                        break;
                }
            }
            if (!saveMode.HasValue)
            {
                if (IsSigned)
                    saveMode = SaveMode.IncrementalUpdate;
            }
            return saveMode.HasValue ? saveMode.Value : SaveMode.Default;

        }

        /// <summary>
        /// Can be null. Save settings passed from client
        /// </summary>
        public virtual SaveSettings GetSaveSettings(string correlationId)
        {

            var opts = GcPdfViewerController.FindDocumentOptions(ClientId, correlationId);
            if (opts != null)
            {
                var saveSettings = opts.saveSettings;
                if (saveSettings != null && saveSettings.sign != null)
                {
                    saveSettings.sign.signerName = saveSettings.sign.signerName.Base64ToUtf8();
                    saveSettings.sign.contactInfo = saveSettings.sign.contactInfo.Base64ToUtf8();
                    saveSettings.sign.location = saveSettings.sign.location.Base64ToUtf8();
                    saveSettings.sign.reason = saveSettings.sign.reason.Base64ToUtf8();
                    saveSettings.sign.signatureField = saveSettings.sign.signatureField.Base64ToUtf8();
                }
                return saveSettings;
            }
            return null;

        }


        /// <summary>
        /// Can be null. Gets the signature settings sent from client. If specified, the document will be saved with signature.
        /// </summary>
        public virtual SignatureInfo GetSignatureInfo(string correlationId)
        {

            var saveSettings = GetSaveSettings(correlationId);
            if (saveSettings != null)
            {
                return saveSettings.sign;
            }
            return null;

        }

        /// <summary>
        /// Save document to memory stream and rewind stream to the beginning position.
        /// </summary>
        /// <param name="s"></param>
        /// <param name="correlationId"></param>
        public virtual void Save(MemoryStream s, string correlationId)
        {
            _applyModifications(correlationId);
            if (GetSignatureInfo(correlationId) != null)
            {
                _saveWithSignature(s, correlationId);
            }
            else
            {
                _doc.Save(s, GetSaveMode(correlationId));
            }
            s.Seek(0, SeekOrigin.Begin);
        }

        /// <summary>
        /// Saves the pages of the current PDF document as PNG images.
        /// </summary>
        /// <returns></returns>
        public virtual List<MemoryStream> SaveAsImages(string correlationId)
        {
            _applyModifications(correlationId);
            var pagesCount = _doc.Pages.Count;
            List<MemoryStream> images = new List<MemoryStream>();
            bool error = false;
            ViewState vs = new ViewState(_doc);
            var saveSettings = GetSaveSettings(correlationId);
            if (saveSettings != null)
            {
                if(saveSettings.zoomFactor > 0)
                {
                    vs.Zoom = saveSettings.zoomFactor;
                }
            }
            _doc.SaveAsImage(vs, Common.OutputRange.All, new SaveAsImageOptions() { DrawAnnotations = true, IgnoreErrors = true }, 
                (GcBitmap bitmap, int pageIndex, int index) =>
            {
                try
                {
                    MemoryStream ms = new MemoryStream();
                    bitmap.SaveAsPng(ms);
                    ms.Seek(0, SeekOrigin.Begin);
                    images.Add(ms);
                }
                catch (Exception)
                {
                    error = true;
#if DEBUG
                    throw;
#endif
                }
            });
            return images;
        }

        public virtual void SaveWithoutModifications(MemoryStream s, string knownFileName)
        {
            _OpenInitialDocState(DocumentId, knownFileName);// Reset previous changes.
            _doc.Save(s);
            s.Seek(0, SeekOrigin.Begin);
        }

        public virtual bool VerifySignature(string fieldName)
        {
            var signatureField = _doc.AcroForm.Fields.Where(f => f.Name == fieldName).FirstOrDefault() as SignatureField;
            if (signatureField == null || signatureField.Value == null)
            {
                return false;
            }
            return signatureField.Value.VerifySignatureValue();
        }

        #endregion

        #region ** IDisposable

        protected virtual void Dispose(bool disposing)
        {
            _doc = null;
            if (Stream != null)
            {
                Stream.Dispose();
                Stream = null;
            }

        }

        public void Dispose()
        {
            // Dispose of unmanaged resources.
            Dispose(true);
            // Suppress finalization.
            GC.SuppressFinalize(this);
        }

        #endregion

        #region ** private and internal implementation

        public ConcurrentDictionary<string, byte[]> AttachedFiles { get; set; } = new ConcurrentDictionary<string, byte[]>();

        private void _applyStructureModifications()
        {
            // Apply document structure changes:                
            var structureChanges = DocumentModifications.structureChanges;
            if (structureChanges != null)
            {
                bool add = false;
                int pageIndex = -1;
                int operationIndex = 0;
                int checkNumPages = -1;

                foreach (var change in structureChanges)
                {
                    add = change.add;
                    pageIndex = change.pageIndex;
                    checkNumPages = change.checkNumPages;
                    pageIndex = Math.Min(pageIndex, checkNumPages); // DOC-2299
                    if (checkNumPages > 5 && !_doc.IsLicensed)
                    {
                        throw new NotLicensedException(GcPdfViewerController.Settings.ErrorMessages.GcPdfNotLicensedLimit);
                    }
                    try
                    {
                        if (checkNumPages != _doc.Pages.Count)
                        {
                            // The most likely reason for this is using an unlicensed GcPdfDocument, which limits the number of pages.
                            // To ignore this error and return whatever pages have been added, comment out the next line:
                            throw new Exception(GcPdfViewerController.Settings.ErrorMessages.NumPagesCheckMismatch);
                        }
                        Page page = null;
                        if (!change.modOnly.HasValue || !change.modOnly.Value)
                        {
                            if (change.add)
                            {
                                if (pageIndex >= _doc.Pages.Count)
                                    page = _doc.Pages.Add();
                                else
                                    page = _doc.Pages.Insert(pageIndex);
                            }
                            else
                            {
                                _doc.Pages.RemoveAt(pageIndex);
                            }
                        }
                        else
                        {
                            page = _doc.Pages[pageIndex];
                        }
                        var pageModification = change.mod;
                        if (pageModification != null)
                        {
                            if (page != null)
                            {
                                if (pageModification.tabs != null)
                                {
                                    if (pageModification.tabs == string.Empty)
                                    {
                                        page.AnnotationsTabsOrder = null; // DOC-3968
                                    }
                                    else
                                    {
                                        page.AnnotationsTabsOrder = DataParser.ParseAnnotationsTabsOrder(pageModification.tabs);
                                    }
                                }
                                if (pageModification.width.HasValue)
                                {
                                    page.Size = new _Size(pageModification.width.Value, page.Size.Height);
                                }
                                if (pageModification.height.HasValue)
                                {
                                    page.Size = new _Size(page.Size.Width, pageModification.height.Value);
                                }
                                if (pageModification.rotate.HasValue)
                                {
                                    page.Rotate = pageModification.rotate.Value;
                                }
                            }
                        }
                        operationIndex++;
                    }
                    catch (Exception)
                    {
                        throw new Exception(GcPdfViewerController.Settings.ErrorMessages.InvalidDocumentStructureChanges);
                        // Check: {operationIndex} {pageIndex} {add} {checkNumPages} {_doc.Pages.Count}. {ex.Message}"
                    }
                }

            }

        }


        private void _redactPdfContent(List<DataParser> annotationsToUpdate, List<DataParser> annotationsToCreate, out List<RedactAnnotation> redactsToApply)
        {
            redactsToApply = new List<RedactAnnotation>();
            Dictionary<PdfObjID, bool> existentAnnotationsRedactedHash = new Dictionary<PdfObjID, bool>();
            foreach (var dataParser in annotationsToUpdate)
            {
                PdfObjID objID = dataParser.pdfObjID;
                AnnotationTypeCode annotationType = dataParser.AnnotationType;
                try
                {
                    if (!dataParser.IsWidget)
                    {
                        AnnotationBase annotation = dataParser.page.Annotations.Where(a => a.ParsedObjID == objID).FirstOrDefault();
                        if (annotation != null)
                        {
                            if (dataParser.redacted && !annotation.ParsedObjID.IsEmpty())
                            {
                                existentAnnotationsRedactedHash[annotation.ParsedObjID] = true;
                            }
                            if (annotationType == AnnotationTypeCode.REDACT && dataParser.ParseBool("redactApplied"))
                            {
                                AnnotationFactory.UpdateAnnotation(annotationType, annotation, dataParser);
                                AddAnnotation(annotation, dataParser.page.Annotations, dataParser);
                                redactsToApply.Add(annotation as RedactAnnotation);
                            }
                        }

                    }
                }
                catch (Exception)
                {
                    continue;
                }
            }
            foreach (var dataParser in annotationsToCreate)
            {
                try
                {
                    var page = dataParser.page;
                    if (!dataParser.IsWidget)
                    {

                        if (dataParser.AnnotationType == AnnotationTypeCode.REDACT && dataParser.redactApplied)
                        {
                            AnnotationBase annotation = AnnotationFactory.CreateAnnotation(dataParser, page);
                            if (annotation != null)
                            {
                                AddAnnotation(annotation, page.Annotations, dataParser);
                                redactsToApply.Add(annotation as RedactAnnotation);
                            }
                        }
                    }
                }
                catch (Exception)
                {
                    continue;
                }
            }

            // Apply all redacts at once  (DOC-2254),
            // keep existent annotations (DOC-4720):
            if (redactsToApply.Count > 0)
            {
                List<KeyValuePair<int, AnnotationBase[]>> keepAnnotationsByPage = new List<KeyValuePair<int, AnnotationBase[]>>();
                foreach (var redact in redactsToApply)
                {
                    if (redact.Page != null && keepAnnotationsByPage.Where(p => p.Key == redact.Page.Index).Count() == 0)
                        keepAnnotationsByPage.Add(new KeyValuePair<int, AnnotationBase[]>(redact.Page.Index, redact.Page.Annotations.ToArray()));
                }
                _doc.Redact(redactsToApply, _settings.RedactOptions);
                foreach (var kv in keepAnnotationsByPage)
                {
                    var page = _doc.Pages[kv.Key];
                    var keepAnnotations = kv.Value;
                    var annotationsRemovedAfterRedact = keepAnnotations.Where(a => !page.Annotations.Any(ina => ina.ParsedObjID == a.ParsedObjID)).ToList();
                    // Restore annotations that should not be deleted
                    foreach (var a in annotationsRemovedAfterRedact)
                    {
                        if (!(a is RedactAnnotation))
                        {
                            if (!a.ParsedObjID.IsEmpty()) // annotation was loaded initially by GcPdf // DOC-4720
                            {
                                if (existentAnnotationsRedactedHash.ContainsKey(a.ParsedObjID))
                                    continue;
                                page.Annotations.Add(a);
                            }
                        }
                    }
                }
            }

        }


        private void _removeRedactedAnnotations(List<DataParser> annotationsToUpdate, List<DataParser> annotationsToCreate)
        {
            // Remove redacted annotations (DOC-4720)
            foreach (var dataParser in annotationsToCreate)
            {
                if (dataParser.redacted)
                    _RemoveAnnotation(dataParser.Annotation, dataParser.pdfObjID);
            }
            foreach (var dataParser in annotationsToUpdate)
            {
                if (dataParser.redacted)
                    _RemoveAnnotation(dataParser.Annotation, dataParser.pdfObjID);
            }
        }
        private void _applyAnnotationModifications(List<DataParser> annotationsToUpdate,
                List<KeyValuePair<int, PdfObjID>> annotationToRemove,
                List<DataParser> annotationsToCreate, out List<AnnotationBase> convertAnnotationsToContent,
                out List<KeyValuePair<PopupAnnotation, string>> popupAnnotations,
                out List<KeyValuePair<MarkupAnnotation, string>> referencedAnnotations)
        {

            convertAnnotationsToContent = new List<AnnotationBase>();
            popupAnnotations = new List<KeyValuePair<PopupAnnotation, string>>();
            referencedAnnotations = new List<KeyValuePair<MarkupAnnotation, string>>();

            foreach (var dataParser in annotationsToUpdate)
            {
                if (dataParser.redacted || dataParser.AnnotationType == AnnotationTypeCode.REDACT)
                    continue;
                PdfObjID objID = dataParser.pdfObjID;
                AnnotationTypeCode annotationType = dataParser.AnnotationType;
                try
                {
                    if (dataParser.IsWidget)
                    {
                        Field field = _FindFieldByObjId(_doc.AcroForm.Fields, objID);
                        if (field == null)
                        {
                            if (dataParser.IsRadioButton)
                                field = _doc.AcroForm.Fields.Where(a => a.Widgets.Any(w => w.ParsedObjID == objID)).FirstOrDefault();
                            if (field == null)
                                continue;
                        }
                        if (!IsFieldNamesEqual(dataParser.FieldName, field))
                        {
                            // DOC-2025
                            var fieldWidget = field.Widgets.Where(fw => fw.ParsedObjID == objID).FirstOrDefault();
                            if (fieldWidget != null)
                            {
                                if (field.Widgets.Count <= 1)
                                {
                                    _RemoveField(field);
                                }
                                else
                                {
                                    _RemoveAnnotation(fieldWidget);
                                }
                                annotationsToCreate.Add(dataParser);
                            }
                            continue;
                        }
                        AcroFormFactory.UpdateField(annotationType, field, dataParser);
                        AddField(field, dataParser);
                        if (dataParser.ConvertToContent)
                        {
                            convertAnnotationsToContent.AddRange(field.Widgets);
                        }
                    }
                    else
                    {
                        AnnotationBase annotation = dataParser.page.Annotations.Where(a => a.ParsedObjID == objID).FirstOrDefault();
                        if (annotation == null)
                        {
                            //throw new Exception($"Annotation with id {objID} not found.");
                            continue;
                        }
                        AnnotationFactory.UpdateAnnotation(annotationType, annotation, dataParser);
                        AddAnnotation(annotation, dataParser.page.Annotations, dataParser);
                        if (dataParser.ConvertToContent)
                        {
                            convertAnnotationsToContent.Add(annotation);
                        }
                        if (annotationType == AnnotationTypeCode.POPUP)
                        {
                            popupAnnotations.Add(new KeyValuePair<PopupAnnotation, string>((PopupAnnotation)annotation, dataParser.parentId));
                        }
                        if ((annotation is MarkupAnnotation) && !string.IsNullOrEmpty(dataParser.referenceAnnotationId))
                        {
                            referencedAnnotations.Add(new KeyValuePair<MarkupAnnotation, string>((MarkupAnnotation)annotation, dataParser.referenceAnnotationId));
                        }
                    }
                }
                catch (Exception)
                {
                    continue;
                }
            }

            // Remove annotations:
            foreach (var removedAnnotation in annotationToRemove)
            {
                try
                {
                    Page page = _doc.Pages[removedAnnotation.Key];
                    PdfObjID objID = removedAnnotation.Value;

                    var field = _doc.AcroForm.Fields.Where(a => a.ParsedObjID == objID).FirstOrDefault();
                    if (field != null)
                    {
                        _RemoveField(field);
                    }
                    else
                    {
                        var annotation = page.Annotations.Where(a => a.ParsedObjID == objID).FirstOrDefault();
                        if (annotation != null)
                        {
                            _RemoveAnnotation(annotation);

                        }

                    }
                }
                catch (Exception)
                {
                    continue;
                }
            }

            // Add new annotations:
            foreach (var dataParser in annotationsToCreate)
            {
                try
                {
                    if (dataParser.redacted || dataParser.AnnotationType == AnnotationTypeCode.REDACT && dataParser.redactApplied)
                        continue;
                    var page = dataParser.page;
                    if (dataParser.IsWidget)
                    {
                        if (dataParser.IsRadioButton || dataParser.IsCheckBox)
                        {
                            string fieldName = dataParser.FieldName;
                            var fieldWithSameName = _doc.AcroForm.Fields.Where(a => a.Name == fieldName).FirstOrDefault();
                            if (fieldWithSameName != null)
                            {
                                var widget = new WidgetAnnotation(dataParser.page, dataParser.rect);
                                fieldWithSameName.Widgets.Add(widget);
                                AcroFormFactory.UpdateFieldWidget(fieldWithSameName, widget, dataParser);
                                if (dataParser.ConvertToContent)
                                    convertAnnotationsToContent.Add(widget);
                                continue;
                            }
                        }
                        Field field = AcroFormFactory.CreateField(dataParser, page);
                        if (field != null)
                        {
                            AddField(field, dataParser);
                            if (dataParser.ConvertToContent)
                                convertAnnotationsToContent.AddRange(field.Widgets);
                        }
                    }
                    else
                    {
                        AnnotationBase annotation = AnnotationFactory.CreateAnnotation(dataParser, page);
                        if (annotation != null)
                        {
                            AddAnnotation(annotation, page.Annotations, dataParser);
                            if (dataParser.ConvertToContent)
                            {
                                convertAnnotationsToContent.Add(annotation);
                            }
                            if (dataParser.AnnotationType == AnnotationTypeCode.POPUP)
                            {
                                popupAnnotations.Add(new KeyValuePair<PopupAnnotation, string>((PopupAnnotation)annotation, dataParser.parentId));
                            }
                        }
                        if ((annotation is MarkupAnnotation) && !string.IsNullOrEmpty(dataParser.referenceAnnotationId))
                        {
                            referencedAnnotations.Add(new KeyValuePair<MarkupAnnotation, string>((MarkupAnnotation)annotation, dataParser.referenceAnnotationId));
                        }
                    }
                }
                catch (Exception)
                {
                    continue;
                }
            }


        }


        private void _applyModifications(string correlationId)
        {
            _OpenInitialDocState(DocumentId, FileName);// Reset previous changes.
            List<DataParser> allClientAnnotations = null;
            List<AnnotationBase> convertAnnotationsToContent = null;
            if (DocumentModifications != null)
            {
                var layersDefaultConfig = _doc.OptionalContent?.DefaultConfig;
                if (layersDefaultConfig != null)
                {
                    var optionalContentVisibility = DocumentModifications.optionalContentVisibility;
                    if (optionalContentVisibility != null)
                    {
                        foreach (var gr in optionalContentVisibility)
                        {
                            if (gr.TryGetValue("groupName", out dynamic groupName) && gr.TryGetValue("visible", out dynamic visible))
                            {
                                try
                                {
                                    layersDefaultConfig.OptionalContent.SetLayerDefaultState(groupName, visible);
                                    //DOC-4067:
                                    layersDefaultConfig.OptionalContent.SetLayerInitialViewState(groupName, visible ? LayerInitialViewState.Always : LayerInitialViewState.Never);
                                }
#if DEBUG
                                catch (Exception ex)
                                {
                                    Console.Error.WriteLine(ex.Message);
                                }
#else
                                catch (Exception) { }
#endif
                            }
                        }
                    }
                }
                if (DocumentModifications.rotation.HasValue)
                    Rotate = DocumentModifications.rotation.Value;
                if (DocumentModifications.renderInteractiveForms.HasValue)
                    RenderInteractiveForms = DocumentModifications.renderInteractiveForms.Value;

                this._applyStructureModifications();

                // Apply annotation modifications
                if (DocumentModifications.annotationsData != null)
                {
                    // Update existing annotations:
                    DataParser.CollectAnnotationsData(out List<DataParser> annotationsToUpdate, out List<DataParser> annotationsToCreate,
                        out List<KeyValuePair<int, PdfObjID>> annotationToRemove, out allClientAnnotations, DocumentModifications, _doc, Info, AttachedFiles);

                    // Apply Redact annotations only for content
                    List<RedactAnnotation> appliedRedacts;
                    _redactPdfContent(annotationsToUpdate, annotationsToCreate, out appliedRedacts);

                    // Apply annotation modifications:
                    List<KeyValuePair<PopupAnnotation, string>> popupAnnotations;
                    List<KeyValuePair<MarkupAnnotation, string>> referencedAnnotations;
                    _applyAnnotationModifications(annotationsToUpdate, annotationToRemove, annotationsToCreate, out convertAnnotationsToContent, out popupAnnotations, out referencedAnnotations);

                    // Remove redacted annotations (DOC-4720)
                    _removeRedactedAnnotations(annotationsToCreate, annotationsToUpdate);

                    // Associate popup annotations with their destinations:
                    foreach (var keyValue in popupAnnotations)
                    {
                        try
                        {
                            var popupAnnotation = keyValue.Key;
                            var parentId = keyValue.Value;
                            var parentObjId = DataParser.IdToPdfObjID(parentId);
                            MarkupAnnotation popupParent = popupAnnotation.Page.Annotations.Where(a => a.ParsedObjID == parentObjId).FirstOrDefault() as MarkupAnnotation;
                            if (popupParent == null)
                            {
                                var annotationParser = allClientAnnotations.Where(p => p.id == parentId).FirstOrDefault();
                                if (annotationParser != null)
                                    popupParent = annotationParser.Annotation as MarkupAnnotation;
                            }
                            if (popupParent != null)
                                popupParent.Popup = popupAnnotation;
                        }
                        catch (Exception)
                        {
                            continue;
                        }
                    }

                    // Set text annotation references:
                    foreach (var keyValue in referencedAnnotations)
                    {
                        try
                        {
                            var referenceAnnotation = keyValue.Key;
                            var referenceAnnotationId = keyValue.Value;
                            var referenceAnnotationObjId = DataParser.IdToPdfObjID(referenceAnnotationId);
                            MarkupAnnotation ownerAnnotation = referenceAnnotation.Page.Annotations.Where(a => a.ParsedObjID == referenceAnnotationObjId).FirstOrDefault() as MarkupAnnotation;
                            if (ownerAnnotation == null)
                            {
                                var annotationParser = allClientAnnotations.Where(p => p.id == referenceAnnotationId).FirstOrDefault();
                                if (annotationParser != null)
                                    ownerAnnotation = annotationParser.Annotation as MarkupAnnotation;
                            }
                            if (ownerAnnotation != null)
                            {
                                referenceAnnotation.ReferenceAnnotation = ownerAnnotation;
                                var reviewOrReplyAnnotation = referenceAnnotation as TextAnnotation;
                                if (reviewOrReplyAnnotation != null)
                                {
                                    if (reviewOrReplyAnnotation.StateModel == "Review")
                                    {
                                        // Review annotation
                                        reviewOrReplyAnnotation.Subject = ownerAnnotation.Subject;
                                        reviewOrReplyAnnotation.ReferenceType = AnnotationReferenceType.Reply;
                                        DataParser.SetFlag(reviewOrReplyAnnotation, AnnotationFlags.Hidden);
                                        DataParser.SetFlag(reviewOrReplyAnnotation, AnnotationFlags.Print);
                                        DataParser.SetFlag(reviewOrReplyAnnotation, AnnotationFlags.NoZoom);
                                        DataParser.SetFlag(reviewOrReplyAnnotation, AnnotationFlags.NoRotate);
                                        reviewOrReplyAnnotation.CreationDate = new PdfDateTime(DateTime.Now);
                                        reviewOrReplyAnnotation.Rect = ownerAnnotation.Rect;
                                        if (string.IsNullOrEmpty(reviewOrReplyAnnotation.Name))
                                        {
                                            reviewOrReplyAnnotation.Name = Guid.NewGuid().ToString();
                                        }
                                        if (reviewOrReplyAnnotation.StateModel == "Review")
                                        {
                                            reviewOrReplyAnnotation.Contents = $"{reviewOrReplyAnnotation.State} set by {reviewOrReplyAnnotation.UserName}";
                                            reviewOrReplyAnnotation.RichText = $"<?xml version=\"1.0\"?><body xmlns=\"http://www.w3.org/1999/xhtml\" xmlns:xfa=\"http://www.xfa.org/schema/xfa-data/1.0/\" xfa:APIVersion=\"Acrobat:20.6.0\" xfa:spec=\"2.0.2\" ><p>{reviewOrReplyAnnotation.Contents}</p></body>";
                                        }
                                    }
                                    else if (reviewOrReplyAnnotation.ReferenceType == AnnotationReferenceType.Reply)
                                    {
                                        if (string.IsNullOrEmpty(reviewOrReplyAnnotation.StateModel))
                                        {
                                            // Reply annotation:
                                            reviewOrReplyAnnotation.CreationDate = new PdfDateTime(DateTime.Now);
                                            DataParser.SetFlag(reviewOrReplyAnnotation, AnnotationFlags.Print);
                                            DataParser.SetFlag(reviewOrReplyAnnotation, AnnotationFlags.NoZoom);
                                            DataParser.SetFlag(reviewOrReplyAnnotation, AnnotationFlags.NoRotate);
                                        }
                                    }
                                }
                            }
                        }
                        catch (Exception)
                        {
                            continue;
                        }
                    }

                }
                // Import form data
                if (DocumentModifications.formData != null && DocumentModifications.formData.Count > 0)
                {
                    try
                    {
                        List<KeyValuePair<string, IList<string>>> fieldValues = new List<KeyValuePair<string, IList<string>>>();
                        foreach (var pageIndex in DocumentModifications.formData.Keys)
                        {
                            if (DocumentModifications.formData.TryGetValue(pageIndex, out Dictionary<string, string[]> pageData))
                            {
                                foreach (var kv in pageData)
                                {
                                    fieldValues.Add(new KeyValuePair<string, IList<string>>(kv.Key, new List<string>(kv.Value)));
                                }
                            }
                        }
                        _doc.ImportFormDataFromCollection(fieldValues.ToArray());
                    }
                    catch (Exception)
                    {
                        //
                    }
                }
            }

            var annotationsOrderTable = DocumentModifications?.annotationsOrderTable;
            int rotate = Rotate;
            var pages = _doc.Pages;
            List<int> pageRotations = new List<int>();
            for (var i = 0; i < pages.Count; i++)
            {
                var page = pages.Where(p => p.Index == i).First();
                var totalRotate = page.Rotate + rotate;
                if (totalRotate > 360)
                    totalRotate = totalRotate - 360;
                pageRotations.Add(totalRotate);
            }
            foreach (var page in pages)
            {
                page.Rotate = pageRotations[page.Index];
                if (annotationsOrderTable != null && annotationsOrderTable.ContainsKey(page.Index))
                {
                    _SortPageAnnotations(page, new List<string>(annotationsOrderTable[page.Index]), allClientAnnotations);
                }
            }
            // Apply stamp rotations:
            if (allClientAnnotations != null)
            {
                foreach (var dataParser in allClientAnnotations)
                {
                    if (dataParser.Annotation != null && dataParser.Annotation.Page != null && dataParser.Annotation is StampAnnotation)
                    {
                        StampAnnotation stamp = (StampAnnotation)dataParser.Annotation;
                        float rotateAngle = dataParser.ParseNumber("rotate", 0);
                        if (rotateAngle != 0)
                        {
                            stamp.RotateContent(rotateAngle);
                        }
                        else
                        {
                            if (stamp.RotationAngle != 0)
                                stamp.RotateContent(0);
                        }

                    }
                }
            }

            // Convert annotations to content
            if (convertAnnotationsToContent != null)
                AnnotationFactory.ConvertAnnotationsToContent(convertAnnotationsToContent);

            // Apply pages filter passed from client
            ApplyPagesFilter(correlationId);
        }

        /// <summary>
        /// Apply pages filter passed from client using save settings.
        /// </summary>
        public virtual void ApplyPagesFilter(string correlationId)
        {
            var saveSettings = GetSaveSettings(correlationId);
            if (saveSettings != null && !string.IsNullOrEmpty(saveSettings.pages))
            {
                try
                {
                    // DOC-5194 >>
                    int? initialGoToPageIndex = null;
                    ActionGoTo initialOpenAction = this._doc.OpenAction as ActionGoTo;
                    if (initialOpenAction != null)
                    {
                        Destination initialOpenActionDest = initialOpenAction.Dest as Destination;
                        if (initialOpenActionDest != null)
                            initialGoToPageIndex = initialOpenActionDest.Page.Index;
                    }
                    // << DOC-5194

                    string[] sFilterSteps = saveSettings.pages.Split(new char[] { ' ', ',' });
                    if (sFilterSteps.Length > 0)
                    {
                        List<PagesFilterStep> resultPages = new List<PagesFilterStep>();
                        for (var stepIndex = 0; stepIndex < sFilterSteps.Length; stepIndex++)
                        {
                            string sFilterStep = sFilterSteps[stepIndex].Trim();
                            if (string.IsNullOrEmpty(sFilterStep))
                                continue;

                            var stepDoc = this._doc;
                            int stepAngle = 0;

                            foreach (Match m in Regex.Matches(sFilterStep, @"\[([^\]]*?)\]", RegexOptions.Multiline))
                            {
                                if (m.Groups.Count > 1)
                                {
                                    var configValue = m.Groups[1].Value;
                                    if (!string.IsNullOrEmpty(configValue))
                                    {
                                        if (configValue.StartsWith("angle:"))
                                        {
                                            stepAngle = int.Parse(configValue.Substring(configValue.IndexOf(":") + 1));
                                        }
                                        if (configValue.StartsWith("file:"))
                                        {
                                            var fileId = configValue.Substring(configValue.IndexOf(":") + 1);
                                            if (fileId == "blank" || fileId == "empty")
                                            {
                                                stepDoc = new GcPdfDocument();
                                            }
                                            else if (AttachedFiles.TryGetValue(fileId, out var bytes))
                                            {
                                                stepDoc = new GcPdfDocument();
                                                stepDoc.Load(new MemoryStream(bytes));
                                            }
                                            else
                                            {
                                                throw new Exception(string.Format(GcPdfViewerController.Settings.ErrorMessages.CannotFindAttachedFileFormat, fileId));
                                            }
                                        }
                                    }
                                }

                            }
                            var lastSq = sFilterStep.LastIndexOf("]");
                            if (lastSq != -1)
                            {
                                sFilterStep = sFilterStep.Substring(lastSq + 1);
                            }
                            if (sFilterStep.Contains("-"))
                            {
                                // Pages range
                                var sArr = sFilterStep.Split('-');
                                int startPageIndex = int.Parse(sArr[0].Trim());
                                int endPageIndex = int.Parse(sArr[1].Trim());
                                for (int pageIndex = startPageIndex; pageIndex <= endPageIndex; pageIndex++)
                                {
                                    while (pageIndex >= stepDoc.Pages.Count)
                                    {
                                        // create empty pages for blank document in case when the fileId has "empty" or "blank" value.
                                        stepDoc.Pages.Add();
                                    }
                                    Page stepDocPage = stepDoc.Pages[pageIndex];
                                    resultPages.Add(new PagesFilterStep(stepDocPage, stepDoc, stepAngle));
                                }
                            }
                            else
                            {
                                // Single page
                                int pageIndex = int.Parse(sFilterStep);
                                resultPages.Add(new PagesFilterStep(stepDoc.Pages[pageIndex], stepDoc, stepAngle));
                            }
                        }
                        _doc.Pages.Clear();
                        for (int i = 0; i < resultPages.Count; i++)
                        {
                            var step = resultPages[i];
                            Page stepPage = step.Page;
                            GcPdfDocument stepDoc = step.Doc;
                            if (stepPage.Owner != null || stepDoc != _doc)
                            {
                                try
                                {
                                    _doc.MergeWithDocument(stepDoc, new MergeDocumentOptions()
                                    {
                                        PagesRange = new OutputRange(stepPage.Index + 1, stepPage.Index + 1)
                                    });
                                    if (step.Angle != 0)
                                        _doc.Pages[_doc.Pages.Count - 1].Rotate += step.Angle;
                                }
                                catch (Exception)
                                {
#if DEBUG
                                    throw;
#endif
                                }
                            }
                            else
                            {
                                if (step.Angle != 0)
                                    stepPage.Rotate += step.Angle;
                                _doc.Pages.Add(stepPage);
                            }
                        }
                    }

                    // DOC-5194 >>
                    if (initialGoToPageIndex.HasValue)
                    {
                        ActionGoTo openAction = this._doc.OpenAction as ActionGoTo;
                        if (openAction != null)
                        {
                            Destination openActionDest = openAction.Dest as Destination;
                            if (openActionDest != null && openActionDest.Page != null && initialGoToPageIndex.Value != openActionDest.Page.Index)
                            {
                                openAction.Dest = null;
                            }
                        }
                    }
                    // << DOC-5194
                }
                catch (Exception ex)
                {
                    var msg = string.Format(GcPdfViewerController.Settings.ErrorMessages.PagesFilterErrorFormat, ex.Message);
                    GcPdfViewerController.SetLastError(msg);
                    throw new Exception(msg, ex);
                }
            }

        }

        private bool IsFieldNamesEqual(string fieldName, Field field)
        {
            if (fieldName == field.Name)
                return true;
            string compoundFieldName = "";
            while (field != null)
            {
                if (string.IsNullOrEmpty(compoundFieldName))
                    compoundFieldName = field.Name;
                else
                    compoundFieldName = $"{field.Name}.{compoundFieldName}";
                field = field.Parent;
            }
            return fieldName == compoundFieldName;
        }

        private Field _FindFieldByObjId(FieldCollection fields, PdfObjID objID)
        {
            if (fields == null)
                return null;
            foreach (var field in fields)
            {
                if (field.ParsedObjID.Equals(objID))
                {
                    return field;
                }
                if (field.Widgets.Where(w => w.ParsedObjID.Equals(objID)).Count() > 0)
                {
                    return field;
                }
                var child = _FindFieldByObjId(field.Children, objID);
                if (child != null)
                    return child;
            }
            return null;
        }

        private Field _FindFieldByName(FieldCollection fields, string fieldName)
        {
            if (fields == null)
                return null;
            foreach (var field in fields)
            {
                if (!string.IsNullOrEmpty(field.Name) && field.Name.Equals(fieldName))
                    return field;
                var child = _FindFieldByName(field.Children, fieldName);
                if (child != null)
                    return child;
            }
            return null;
        }

        private string _FindClientId(AnnotationBase annotation, List<DataParser> allClientAnnotations)
        {
            if (annotation == null)
                return "";
            string clientId = annotation.ParsedObjID.IsEmpty() ? _FindAnnotationId(annotation, allClientAnnotations) : DataParser.PdfObjIDtoId(annotation.ParsedObjID);
            if (string.IsNullOrEmpty(clientId) && annotation is WidgetAnnotation)
            {
                var dataParser = allClientAnnotations.Where(a => a.IsWidget == true && a.Widget != null && a.Widget == annotation).FirstOrDefault();
                if (dataParser == null)
                {
                    var field = (annotation as WidgetAnnotation).Field;
                    dataParser = allClientAnnotations.Where(a => a.IsWidget == true && a.Field == field).FirstOrDefault();
                }
                if (dataParser != null)
                {
                    clientId = dataParser.id;
                }
            }
            return clientId;
        }

        private void _SortPageAnnotations(Page page, List<string> annotationsOrder, List<DataParser> allClientAnnotations)
        {
            AnnotationCollection annotations = page.Annotations;
            if (annotations == null || annotationsOrder == null || allClientAnnotations == null)
                return;
            try
            {
                annotations.Sort(new Comparison<AnnotationBase>(delegate (AnnotationBase x, AnnotationBase y)
                {
                    if (x == null && y == null)
                        return 0;
                    if (x == null)
                        return -1;
                    if (y == null)
                        return 1;
                    try
                    {
                        string xID = _FindClientId(x, allClientAnnotations);
                        string yID = _FindClientId(y, allClientAnnotations);
                        int xIndex = annotationsOrder.IndexOf(xID);
                        int yIndex = annotationsOrder.IndexOf(yID);
                        if (xIndex == -1 || yIndex == -1)
                        {
                            if (x is WidgetAnnotation widget1 && y is WidgetAnnotation widget2)
                            {
                                var xName = widget1.Field?.Name ?? "";
                                var yName = widget2.Field?.Name ?? "";
                                return xName.CompareTo(yName);
                            }
                            // annotation without id (-1) should go to the end of the list (fix DOC-3358)
                            return yIndex.CompareTo(xIndex);
                        }
                        int res = xIndex.CompareTo(yIndex);
                        return res;
                    }
                    catch (Exception)
                    {
                        return -1;
                    }

                }));
            }
            catch (Exception)
            {
#if DEBUG
                Console.Error.WriteLine("[Debug] [SupportApi] Exception in _SortPageAnnotations");
#endif
            }
        }

        private string _FindAnnotationId(AnnotationBase annotation, IEnumerable<DataParser> allClientAnnotations)
        {
            if (allClientAnnotations == null)
                return "";
            var fieldParser = allClientAnnotations.Where(p => p.Field != null && p.Field.Equals(annotation)).FirstOrDefault();
            return fieldParser != null ? fieldParser.id : "";
        }

        private void AddField(Field field, DataParser dataParser)
        {
            try
            {
                FieldCollection owner = field.Owner;
                string uniqueFieldName = field.Name;
                // Remove original field from collection. (DOC-2001)
                _RemoveField(field);
                dataParser.Field = field;
                // Then search for duplicate names. (DOC-2001)
                int duplicateCount = 0;
                Field duplicateField = _FindFieldByName(_doc.AcroForm.Fields, uniqueFieldName);
                while (duplicateField != null)
                {
                    if (duplicateField.Widgets.Count == 0 || duplicateField.Widgets[0].Page == null)
                    {
                        // Fix for DOC-1991: replace field without page.
                        _RemoveField(duplicateField);
                        break;
                    }
                    uniqueFieldName = $"{field.Name}_{++duplicateCount}";
                    duplicateField = _FindFieldByName(_doc.AcroForm.Fields, uniqueFieldName);
                }
                field.Name = uniqueFieldName;

                if (owner == null)
                    owner = _doc.AcroForm.Fields;
                owner.Add(field);
                foreach (var widget in field.Widgets)
                    widget.Page = dataParser.page;
            }
            catch (Exception)
            {
                throw;
            }
        }

        private void AddAnnotation(AnnotationBase annotation, AnnotationCollection collection, DataParser dataParser)
        {
            _RemoveAnnotation(annotation);
            dataParser.Annotation = annotation;
            collection.Add(annotation);
            annotation.Page = dataParser.page;
        }

        private void _RemoveAnnotation(AnnotationBase annotation, PdfObjID? pdfObjID = null)
        {
            foreach (var p in _doc.Pages)
            {
                if (annotation == null && pdfObjID.HasValue && !pdfObjID.Value.IsEmpty())
                    annotation = p.Annotations.Where(a => a.ParsedObjID == pdfObjID.Value).FirstOrDefault();
                if (annotation != null)
                {
                    if (p.Annotations.Contains(annotation))
                    {
                        p.Annotations.Remove(annotation);
                        break;
                    }
                }
            }
        }

        private void _RemoveField(Field field)
        {
            foreach (var widget in field.Widgets.ToArray())
            {
                _RemoveAnnotation(widget);
            }
            var owner = field.Owner;
            if (owner == null)
                owner = _doc.AcroForm.Fields;
            owner.Remove(field);
        }

        private void _saveWithSignature(Stream stream, string correlationId)
        {
            var _signatureInfo = GetSignatureInfo(correlationId);
            if (_signatureInfo != null)
            {
                var signatureProperties = new SignatureProperties();
                signatureProperties.ContactInfo = _signatureInfo.contactInfo;
                signatureProperties.Location = _signatureInfo.location;
                signatureProperties.SignerName = _signatureInfo.signerName;
                signatureProperties.SigningDateTime = DateTime.Now;
                signatureProperties.Reason = _signatureInfo.reason;
                signatureProperties.SignatureField = _doc.AcroForm.Fields.Where(f => f.Name == _signatureInfo.signatureField).FirstOrDefault();
                var e = new SignEventArgs(signatureProperties, _doc);
                _settings.OnSign(e);
                if (e.Reject)
                {
                    throw new Exception(GcPdfViewerController.Settings.ErrorMessages.DocumentSigningRejected);
                }
                _doc.Sign(signatureProperties, stream, GetSaveMode(correlationId));
            }
        }

        private OpenDocumentInfo _OpenInitialDocState(string knownDocumentId, string knownFileName)
        {
            try
            {
                if (_doc != null)
                {
                    _doc.Clear();
                    _doc = null;
                }
                string password = null;
                if (_docOptions != null)
                    password = _docOptions.password;
                var d = new GcPdfDocument();
                Stream.Seek(0, SeekOrigin.Begin);
                d.Load(Stream, password);
                _doc = d;
                _Size defaultViewPortSize = d.Pages.Count > 0 ? d.Pages[0].GetRenderSize(this._resolution, this._resolution) : new _Size(0f, 0f);

                string fileName = _docOptions?.fileName ?? string.Empty;
                if (_doc.Metadata != null)
                {
                    if (string.IsNullOrEmpty(fileName))
                    {
                        fileName = $"{_doc.Metadata.Title ?? ""}";
                    }
                }
                string documentId = knownDocumentId;
                if (string.IsNullOrEmpty(documentId))
                {
                    // Auto-generate document id:
                    documentId = $"generated_id_{Guid.NewGuid()}".Replace("-", "");
                    //documentId = _doc.Metadata.Identifier; // do not use fingerprint
                }
                if (string.IsNullOrEmpty(fileName))
                {
                    fileName = string.IsNullOrEmpty(knownFileName) ? documentId : knownFileName;
                }
                DocumentId = documentId;
                Info = new OpenDocumentInfo(documentId, fileName, _accessMode, d.DocumentInfo, d.Pages.Count, ClientId, defaultViewPortSize, _docOptions);
                return Info;
            }
            catch (Exception ex)
            {
                GcPdfViewerController.SetLastError(ex.Message);
                throw;
            }
        }

        #endregion

    }

    internal class PagesFilterStep
    {
        private readonly int _AdditionalAngle;
        private readonly int _InitialRotate;

        public PagesFilterStep(Page page, GcPdfDocument doc, int angle = 0)
        {
            this.Page = page;
            this.Doc = doc;
            this._InitialRotate = page.Rotate;
            this._AdditionalAngle = angle;
        }

        public Page Page { get; set; }
        public GcPdfDocument Doc { get; set; }

        public int Angle
        {
            get
            {
                // DOC-5226:
                var angleDiff = this.Page.Rotate - this._InitialRotate;
                return this._AdditionalAngle - angleDiff;

            }
        }
    }


}
